package com.dsk.flink.dsc.sync;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSONObject;
import com.dsk.flink.dsc.common.function.CanalMapToTsGroupFunction;
import com.dsk.flink.dsc.common.function.GroupTsProcessWindowFunction;
import com.dsk.flink.dsc.common.function.MysqlDataTransferFunction;
import com.dsk.flink.dsc.common.sink.MysqlDataTransferSink;
import com.dsk.flink.dsc.utils.EnvProperties;
import com.dsk.flink.dsc.utils.EnvPropertiesUtil;
import com.dsk.flink.dsc.utils.EtlUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.functions.FilterFunction;
import org.apache.flink.api.common.restartstrategy.RestartStrategies;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.java.tuple.Tuple3;
import org.apache.flink.api.java.utils.ParameterTool;
import org.apache.flink.streaming.api.CheckpointingMode;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.CheckpointConfig;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.assigners.TumblingProcessingTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaConsumer;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Comparator;
import java.util.List;

/**
 * @author shezaixing
 * @date 2023/12/5 14:44
 * @description 同步至客户目标数据源任务
 * update by lww
 */
@Slf4j
public class SyncCustomerDataSource {

    public static void main(String[] args) throws Exception {

        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
        //env.setParallelism(1);
        env.setRestartStrategy(RestartStrategies.fixedDelayRestart(100, 60000));
        env.enableCheckpointing(180000);
        env.getCheckpointConfig().setCheckpointingMode(CheckpointingMode.EXACTLY_ONCE);
        env.getCheckpointConfig().setMinPauseBetweenCheckpoints(180000);
        env.getCheckpointConfig().setCheckpointTimeout(300000);
        env.getCheckpointConfig().setMaxConcurrentCheckpoints(1);
        env.getCheckpointConfig().setExternalizedCheckpointCleanup(CheckpointConfig.ExternalizedCheckpointCleanup.RETAIN_ON_CANCELLATION);
        env.getCheckpointConfig().setTolerableCheckpointFailureNumber(100);

        //获取用户自己的配置信息
        ParameterTool parameterTool = ParameterTool.fromArgs(args);
        String offsetTimestamp = parameterTool.get("offsetTimestamp");
        String propertiesPath = parameterTool.get("propertiesPath");
        EnvProperties envProps = EnvPropertiesUtil.getPropertiesFromArgsPath(propertiesPath);
        //envProps.put("providerImpl", JdbcConnectionProviderFactory.HikariDataSourceJdbcConnectionProvider.class.getName());
        System.out.println("读取到的配置文件：-> " + envProps.toString());
        System.out.println("读取到的数据连接配置：->" + String.format(envProps.getDb_url(), envProps.getDb_host(), envProps.getDb_port(), envProps.getDb_database()));
        System.out.println("获取到的kafka消费组：->" + EtlUtils.getKafkaGroup(envProps));
        FlinkKafkaConsumer<String> kafkaConsumer = new FlinkKafkaConsumer<String>(envProps.getKafka_topic(), new SimpleStringSchema(), EtlUtils.getKafkaConfig(envProps.getKafka_brokers(), EtlUtils.getKafkaGroup(envProps), envProps.getKafka_username(),envProps.getKafka_password()));
        //System.out.println(envProps.getKafka_topic());
        long defaultOffset = LocalDateTime.now().minusMinutes(10).atZone(ZoneId.systemDefault()).toInstant().toEpochMilli();
        kafkaConsumer.setStartFromTimestamp(defaultOffset);
        //kafkaConsumer.setStartFromLatest();
        //偏移量
        if (StrUtil.isNotBlank(offsetTimestamp)) {
            kafkaConsumer.setStartFromTimestamp(Long.parseLong(offsetTimestamp));
        }

        SingleOutputStreamOperator<Tuple3<JSONObject, String, Long>> tsGroupStream = env.addSource(kafkaConsumer)
                .map(JSONObject::parseObject)
                .filter(new FilterFunction<JSONObject>() {
                    @Override
                    public boolean filter(JSONObject value) throws Exception {
                        return !value.getBoolean("isDdl") && !"TIDB_WATERMARK".equals(value.getString("type"));
                    }
                })
                .name("dsc-source")
                .uid("dsc-source")
                .map(new CanalMapToTsGroupFunction())
                .name("dsc-groupKey")
                .uid("dsc-groupKey");

        SingleOutputStreamOperator<JSONObject> process = tsGroupStream.keyBy(x -> x.f1)
                .window(TumblingProcessingTimeWindows.of(Time.milliseconds(50)))
                .process(new GroupTsProcessWindowFunction())
                .uid("dsc-w1")
                .name("dsc-w1");

        OutputTag<Tuple3<String, String, Long>> cdcLogTag = new OutputTag<Tuple3<String, String, Long>>("dsc_cdc_log") {};
        SingleOutputStreamOperator<Tuple3<String, String, Long>> slide = process
                .process(new MysqlDataTransferFunction(envProps, cdcLogTag))
                .name("dsc-sql")
                .uid("dsc-sql");

        SingleOutputStreamOperator<String> groupWindowSqlResultStream = slide
                .keyBy(value -> value.f1)
                .window(TumblingProcessingTimeWindows.of(Time.milliseconds(50)))
                .process(new ProcessWindowFunction<Tuple3<String, String, Long>, String, String, TimeWindow>() {
                    @Override
                    public void process(String s, ProcessWindowFunction<Tuple3<String, String, Long>, String, String,
                            TimeWindow>.Context context, Iterable<Tuple3<String, String, Long>> elements,
                                        Collector<String> out) throws Exception {
                        List<Tuple3<String, String, Long>> list = CollUtil.list(false, elements);
                        if (CollUtil.isNotEmpty(list)) {
                            Tuple3<String, String, Long> maxTsElement =
                                    list.stream().max(Comparator.comparing(x -> x.f2)).get();
                            out.collect(maxTsElement.f0);
                        }
                    }
                })
                .name("dsc-max")
                .uid("dsc-max");

        groupWindowSqlResultStream.addSink(new MysqlDataTransferSink(envProps))
                .name("dsc-sink")
                .uid("dsc-sink");

        slide.getSideOutput(cdcLogTag).map(x -> x.f0).addSink(new MysqlDataTransferSink(envProps))
                .name("dsc-cdc-log")
                .uid("dsc-cdc-log");

        env.execute("dsc-client");
    }
}
